declare class X {
  foobar<T>(): void;
  delete<T>(): void;
  yield<T>(): void;
  do<T>(): void;
  static foobar<T>(): void;
  static delete<T>(): void;
  static yield<T>(): void;
  static do<T>(): void;
};
