import {type} from "foo";
import {type t} from "foo";
import {type as} from "foo";
import {type t as u} from "foo";

import {typeof t} from "foo";
import {typeof as} from "foo";
import {typeof t as u} from "foo";
