class A {
  get
  () {}

  set
  () {}

  static
  () {}

  async
  () {}


  static
  get
  () {}

  static
  set
  () {}

  static
  static
  () {}

  static
  async
  () {}

  static
  a
  () {}


  get
  async
  () {}


  static
  get
  static
  () {}
}