class A1 {
  static
  a
  static
}

class A2 { a }
class A3 { get }
class A4 { set }
class A5 { static }
class A6 { async }

class A7 {
  get
  *a(){}
}

class A8 {
  static
  *a(){}
}

class A9 {
  async
  a(){}
}

class A10 {
  static
  async
  a
}

class A11 { static; }

class A12 {
  static = 0;
}

class A13 {
  get
  ['a'](){}
}

class A14 {
  static
  get
  static
  (){}
}