// bounds
type T = any;
type T = mixed;
type T = empty;
// builtins
type T = void;
type T = number;
type T = string;
type T = bool;
type T = boolean;
// literal type annotations
type T = null;
type T = "";
type T = 0;
type T = true;
type T = false;
